#!/usr/bin/env python3

import geojson as gj
import overpass
import logging

from geopandas import GeoDataFrame
from pathlib import Path

from .utils import hash_
from . import defaults

class POIDataset(GeoDataFrame):
    '''
    Extension of a GeoDataFrame which stores Points only, to be used as origin, destination,
    and waypoints in routing.

    Parameters
    ----------
    name: str
        Name of POIDataset (required)
    '''
    def __init__(self, *args, name=None, **kwargs):

        if name is None:
            raise ValueError("Must specify name for POIDataset")
        else:
            setattr(self, "_metadata", {"name": name})

        super(POIDataset, self).__init__(*args, **kwargs)

    def get_name(self):
        """Return POI dataset name"""
        return self._metadata["name"]

    def _filter_points(self, gdf):
        '''Filter out any records in GeoDataFrame that are not Point geometries'''

        gdf_geom_types = gdf.apply(lambda r: r.geometry.type, axis=1)
        gdf_points = gdf[gdf_geom_types == "Point"]

        return gdf_points

    @classmethod
    def from_overpass(cls, query, name=None, overwrite=False, tmp_dir=defaults.TMP_DIR, **kwargs):
        '''
        Initialize POIDataset from Overpass API query. Any returned nodes or closed ways with `center` attributes will be included in the dataset.

        Parameters
        ----------
        query : str
            Query to be sent to overpass API. This query should *not* include an `out` directive (eg. [out:json];)
        name : str
            Name of the POI dataset
        overwrite : bool
            Overwrite POIDataset if it already exists on disk
        tmp_dir : str
            Temporary directory to save POIDataset
        '''
        
        logger = logging.getLogger(defaults.LOGGER)

        ## Use md5 hash of query as filename if name not specified
        out_folder = Path(tmp_dir)
        out_name = name if name else hash_(query)
        out_file = out_folder / "{}.geojson".format(out_name)

        ## Honor overwrite settings
        if out_file.is_file():
            if overwrite:
                logger.info("Overwriting {}".format(out_file))
                out_file.unlink()
            else:
                logger.info("Using existing POIDataset {}".format(out_file))
                return cls.from_file(out_file, name=out_name).set_index("id")


        logger.info("Downloading POIDataset {}".format(name))

        ## Query API
        oapi = overpass.API()
        json = oapi.get("[out:json];{}".format(query),
                        verbosity="geom",
                        responseformat="json",
                        build=False)

        ## Filter response and convert to FeatureCollection
        features = []

        for el in json['elements']:
            ## Only look at ways & nodes
            if not "type" in el or el['type'] not in ["way", "node"]:
                continue

            ## Discard ways without center points
            if el['type'] == "way" and "center" in el:
                feat = gj.Feature(geometry = gj.Point((el['center']['lon'], el['center']['lat'])),
                                  properties = {**{"id": el['id']}, **el['tags']})

            elif el['type'] == "node":
                feat = gj.Feature(geometry = gj.Point((el['lon'], el['lat'])),
                                  properties = {**{"id": el['id']}, **el['tags']})

            else:
                continue

            features.append(feat)

        pois = gj.FeatureCollection(features)

        ## Cache POIs for next time
        out_folder.mkdir(parents=True, exist_ok=True)
        out_file.write_text(gj.dumps(pois))

        return cls.from_features(pois.features, name=out_name, crs={'init': 'epsg:4326'}) \
                  .set_index("id")

    @classmethod
    def from_file(cls, path, name=None, **kwargs):
        '''Initialize POIDataset from file. If no name is given, the filename will be used'''

        gdf = GeoDataFrame.from_file(path, **kwargs)
        gdf_points = cls._filter_points(cls, gdf)

        return cls(gdf_points, name=name if name else Path(path).stem)

    @classmethod
    def from_features(cls, features, name=None, **kwargs):
        '''Initialize POIDataset from GeoJSON features'''

        gdf = GeoDataFrame.from_features(features, **kwargs)
        gdf_points = cls._filter_points(cls, gdf)

        return cls(gdf_points, name=name)



from osrm import AccessIsochrone as OSRMAccessIsochrone
from shapely.geometry import MultiPolygon

class AccessIsochrone(OSRMAccessIsochrone):
    """
    Compute an access isochrone from an origin point with a given `ScenarioAPI`

    Parameters
    ----------
    scenario : Scenario
        Scenario to be queried for access isochrone
    point_origin : 2-floats tuple
        The coordinates of the center point to use as (x, y).
    points_grid : int
        The number of points of the underlying grid to use.
    size : float
        Search radius (in wgs84 degrees)
    """
    def __init__(self, scenario, point_origin, points_grid=500, size=0.4):

        super(AccessIsochrone, self).__init__(point_origin, points_grid, size,
                                              url_config = scenario.config)

    def get_center(self):
        """Return center point used in isochrone calculations"""
        return self.center_point

    def get_grid(self):
        """Return GeoDataFrame of grid used in duration calculations, snapped to the road network"""
        return self.grid

    def get_durations(self):
        """Return durations table retrieved from OSRM"""
        return self.times

    def render_contour(self, n_levels):
        """Return GeoDataFrame of MultiPolygon contours for a specified number of levels """

        contours = super(AccessIsochrone, self).render_contour(n_levels)

        ## osrm.AccessIsochrone.render_contour returns a GeoDataFrame with mixed
        ## Polygon/Multipolygon geometries. This ensures all geometries are multipolygons
        ## and renames the `time` field which is a special column for many data types and
        ## causes issues on export
        contours.geometry = contours.geometry.apply(lambda g: MultiPolygon([g]) if g.type == "Polygon" else g)
        contours.rename(columns={"time": "duration"}, inplace=True)

        return contours



#!/usr/bin/env python3

import logging
import atexit
import sys
import sh

from . import defaults

class OSRM():
    """
    Base class wrapper around all the osrm-* binaries. All functions accept
    additional **kwargs to be passed upon execution. This class is used in
    the background and need not be initialized directly.

    Parameters
    ----------
    verbose: bool
        Output stdout from osrm commands.
    """

    def __init__(self, verbose=defaults.VERBOSE):
        self.log = logging.getLogger(defaults.LOGGER)
        self.verbose = verbose
        self.processes = []

        ## Make sure all OSRM are killed upon script exit
        atexit.register(self._kill_all_osrm)

        try:
            self._osrm_extract = sh.Command("osrm-extract")
            self._osrm_contract = sh.Command("osrm-contract")
            self._osrm_routed = sh.Command("osrm-routed")
            self._osrm_partition = sh.Command("osrm-partition")
            self._osrm_customize = sh.Command("osrm-customize")
            self._osrm_datastore = sh.Command("osrm-datastore")

        except CommandNotFound as err:
            self.log.error("OSRM binary not found: {}".format(err))

    def extract(self, osm_path, profile_path, **kwargs):
        '''
        Call `osrm-extract` with a path to the osm route network and lua profile. 
        '''

        defaults = {"_bg": True,
                    "_out": sys.stdout if self.verbose else None}

        proc = self._osrm_extract(osm_path, "-p", profile_path, **{**defaults, **kwargs})

        self._log_cmd(proc)
        self.processes.append(proc)

        return proc

    def partition(self, osrm_file, **kwargs):
        '''Call `osrm-partition` for a .osrm file'''
      
        defaults = {"_bg": True,
                    "_out": sys.stdout if self.verbose else None}

        proc = self._osrm_partition(osrm_file, **{**defaults, **kwargs})

        self._log_cmd(proc)
        self.processes.append(proc)

        return proc

    def customize(self, osrm_file, **kwargs):
        '''Call `osrm-customize` on a .osrm file'''
      
        defaults = {"_bg": True,
                    "_out": sys.stdout if self.verbose else None}

        proc = self._osrm_customize(osrm_file, **{**defaults, **kwargs})

        self._log_cmd(proc)
        self.processes.append(proc)

        return proc

    def contract(self, osrm_file, **kwargs):
        '''Call `osrm-contract` on a .osrm file'''

        defaults = {"_bg": True,
                    "_out": sys.stdout if self.verbose else None}

        proc = self._osrm_contract(osrm_file, **{**defaults, **kwargs})

        self._log_cmd(proc)
        self.processes.append(proc)

        return proc

    def routed(self, osrm_file, ready_callback, done_callback, verbose=False, **kwargs):
        '''
        Call `osrm-routed` on a .osrm file

        Parameters
        ----------
        osrm_file : str
            Path to *.osrm
        ready_callback : function
            Function to be called when osrm-routed is ready for HTTP requests
        done_callback : function
            Function to be called when osrm-routed has exited
        verbose : bool
            `osrm-routed` output is so verbose it is default off even if the parent class `verbose=True`. Set this to true if you want to see `osrm-routed` output anyway.
        **kwargs
            Any additional parameters to be passed to osrm-routed
        '''

        ## Parse OSRM output line by line and exec callback when server is running
        def parse_output(line, stdin, process):
            if line.find("running and waiting for requests"):
                ready_callback(process)
                return True ## Returning True stops executing this function each line

        def parse_output_verbose(line, stdin, process):
            if line.find("running and waiting for requests"):
                ready_callback(process)

            print(line)

        defaults = {"_bg": True, "_bg_exc": False,
                    "verbosity": "INFO", "ip": "127.0.0.1",
                    "_out": parse_output_verbose if verbose else parse_output,
                    "_done": done_callback}

        proc = self._osrm_routed(osrm_file, **{**defaults, **kwargs})

        self._log_cmd(proc)
        self.processes.append(proc)

        return proc

    def get_version(self):
        '''Return OSRM binaries version'''
        return self._osrm_extract("--version")

    def _log_cmd(self, proc):
        '''Log command text in a debug log output'''
        cmd = " ".join([b.decode('utf-8') for b in proc.cmd])

        self.log.debug("Executing: {}".format(cmd))

    def _kill_all_osrm(self):
        '''
        Send kill signal to all osrm processes spawned under this module. This will
        be executed upon script exit to ensure that there's no processes left behind
        '''

        for proc in self.processes:
            try:
                proc.kill()
            except:
                continue



#!/usr/bin/env python3

import subprocess
import logging
import shutil
import osrm
import time
import sys
import sh

from contextlib import AbstractContextManager
from urllib.error import HTTPError
from sh import CommandNotFound
from functools import partial
from pathlib import Path

from .RoutingProfile import RoutingProfile
from .OSMDataset import OSMDataset
from .OSRM import OSRM
from . import defaults
from . import utils

class Scenario():
    """
    Scenario is an abstraction of OSRM executables in order to compile, serve, and query
    a routable network.

    A Scenario is initialized with (1) an OSM Dataset and (2) a Routing Profile. When called,
    Scenario compiles the OSMDataset and RoutingProfile into an OSRM routable network,
    provides a context manager to serve the scenario's HTTP API, and returns various methods
    to query that API.

    The HTTP API methods (`match`, `nearest`, `simple_route`, `table`, `trip`) are provided by
    the `python-osrm` module (https://github.com/ustroetz/python-osrm)

    Example
    -------
    >>> from tebetebe.profiles import foot
    >>> import tebetebe as tb
    >>>
    >>> ## initialize scenario with GeoFabrik extract & default foot profile
    >>> scenario = tb.Scenario("./swaziland-latest.osm.pbf", foot)
    >>>
    >>> with scenario() as api: ## compile scenario and serve HTTP API
    >>>     api.simple_route()... ## query HTTP API for a simple route, match, nearest, trip...

    Parameters
    ----------
    osm_dataset : str / OSMDataset
        OSM dataset from which the route network will be extracted
    routing_profile : str / RoutingProfile
        Routing profile to be used in scenario
    name : str, optional
        Scenario name used in output file and log entries. If not supplied, it will be built
        based upon the OSMDataset and RoutingProfile names
    algorithm : str, optional
        Algorithm to be used (either "CH" or "MLD")
    tmp_dir : str, optional
        Temporary directory to store files generated by osrm binaries
    overwrite: bool, optional
        Overwrite scenario if .osrm* files already exist. Otherwise, the existing .osrm* files
        will be used.
    verbose: bool, optional
        Print output of OSRM compilation
    **kwargs
        To pass any custom arguments to the OSRM executables ("osrm-routed", "osrm-contract", ...)
        pass the kwarg "{executable}_args" (ex. "routed_args", "contract_args", ...) with a dictionary
        of key:values to be passed
    """

    def __init__(self, osm_dataset, routing_profile,
                 name=None, algorithm="MLD", 
                 tmp_dir=defaults.TMP_DIR,
                 overwrite=defaults.OVERWRITE,
                 verbose=defaults.VERBOSE,
                 **kwargs):

        self.log = logging.getLogger(defaults.LOGGER)
        self.algorithm = algorithm.upper()
        self.tmp_dir = Path(tmp_dir)
        self.overwrite = overwrite

        ## Handle if params are passed as string
        if isinstance(osm_dataset, str):
            osm_dataset = OSMDataset(osm_dataset, tmp_dir=self.tmp_dir)

        if isinstance(routing_profile, str):
            routing_profile = RoutingProfile(routing_profile)

        self.osm_dataset = osm_dataset
        self.routing_profile = routing_profile

        ## Initialize / check OSRM Binaries
        self.OSRM = OSRM(verbose=verbose)

        ## Scenario name and output osrm path
        self.name = name if name else "{}_{}".format(self.osm_dataset.get_name(),
                                                     self.routing_profile.get_name())
        self.path = self.tmp_dir / self.name / "{}.osrm".format(self.name)
        self.path.parent.mkdir(parents=True, exist_ok=True) # Make sure output_dir exists

        ## Add osrm executable args if they are passed
        self.args = {}
        for k_name, kwarg in kwargs.items():
            if k_name[-5:] == "_args":
                self.args[k_name[:-5]] = kwarg

    ##
    ## Scenario compilation
    ##

    def __call__(self):
        '''
        Compile OSMDataset and RoutingProfile into an OSRM routable network (.osrm file)
        '''

        ## Honor overwrite settings
        if self.path.is_file():
            if self.overwrite:
                self.log.info("{}: Overwriting scenario".format(self.name))

                ## Remove all .osrm* files
                for osrm_file in self.path.parent.glob("{}.osrm*".format(self.name)):
                    osrm_file.unlink()
            else:
                self.log.info("{}: Using existing scenario".format(self.name))
                return self

        try:
            self.log.info("{}: Compiling scenario ({})".format(self.name, self.algorithm))

            if self.routing_profile.is_default():
                self.log.warning("{}: Default {} profile may not be accurate for your use case".format(self.name,
                                                                                                       self.routing_profile.get_name()))

            ## Run osrm-extract
            extr = self.OSRM.extract(self.osm_dataset.get_path(),
                                     self.routing_profile.get_path(),
                                     **self._get_command_args("extract"))
            extr.wait() ## synchronous execution

            ## osrm-extract generates osrm file in same folder as OSMDataset
            ## so rename and move all files to self.path
            network_path = self.osm_dataset.get_path()

            for src_path in network_path.parent.glob("{}.osrm*".format(network_path.stem.split(".")[0])):
                dst_file = src_path.name.replace(src_path.stem.split(".")[0], self.name)
                dst_path = self.path.parent / dst_file

                shutil.move(src_path, dst_path)

            ## Run different binaries depending on algorithm used
            if self.algorithm == "MLD":
                part = self.OSRM.partition(self.path, **self._get_command_args("partition"))
                part.wait()

                cust = self.OSRM.customize(self.path, **self._get_command_args("customize"))
                cust.wait()

            elif self.algorithm == "CH":
              cont = self.OSRM.contract(self.path, **self._get_command_args("contract"))
              cont.wait()

            return self

        except Exception as exc:
            ## TODO custom handle more common exception
            self.log.error(exc)

    ##
    ## Context Manager for running Scenario HTTP server (osrm-routed)
    ##

    def __enter__(self):
        ## Compile the Scenario if it hasn't yet
        if not self.path.is_file():
            self()

        ## Point the osrm http api to correct port
        port = utils.find_open_port()
        self._set_config(port)

        self._set_command_args("routed", {
            "port": port,
            "algorithm": self.algorithm,
            "dataset_name": self.name
        })

        self.ready = False

        ## Start up OSRM HTTP server with OSRM-routed
        try:
            ## Callback to execute when OSRM process is ready to receive http requests
            def ready_callback(process):
                self.ready = True

            ## Callback when command exits
            def done_callback(cmd, success, exit_code):
                ## -9 = SIGKILL = it was (probably) killed by this script. aka, this is normal.
                if exit_code != -9:
                    self.log.error("{}: Scenario process exited with exit code {}".format(self.name, exit_code))
                    self.log.error("{}: Command: {}".format(self.name, cmd))


            self.log.info("{}: Initializing scenario".format(self.name))
            self.process = self.OSRM.routed(self.path,
                                            ready_callback,
                                            done_callback,
                                            **self._get_command_args("routed"))

        except Exception as exc:
            self.log.error(exc)

        ## Wait until OSRM-routed is ready before returning the api
        ## TODO better way to do this?
        while not self.ready:
            time.sleep(1)

        self.log.info("{}: Ready for requests".format(self.name))

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        ## TODO handle common exceptions
        if exc_type:
            if exc_type is HTTPError: ## Log but not exit on HTTP Errors
                self.log.error(exc_value)
                return True
            else:
                self.log.error("{} {} {}".format(exc_type, exc_value, traceback))

        ## Kill the OSRM process
        try:
            self.process.kill()
        except:
            pass

    ##
    ## osrm-routed HTTP API

    simple_route = osrm.simple_route
    nearest = osrm.nearest
    match = osrm.match
    table = osrm.table
    trip = osrm.trip

    def _set_config(self, port):
        """Set up HTTP API configuration given the HTTP server port"""
        self.config = osrm.RequestConfig("127.0.0.1:{}/v1/skobuffs".format(port))

        self.simple_route = partial(osrm.simple_route, url_config=self.config)
        self.nearest = partial(osrm.nearest, url_config=self.config)
        self.match = partial(osrm.match, url_config=self.config)
        self.table = partial(osrm.table, url_config=self.config)
        self.trip = partial(osrm.trip, url_config=self.config)

    ##
    ## Other

    def _get_command_args(self, command):
        return self.args[command] if command in self.args else {}

    def _set_command_args(self, command, args):
        self.args[command] = self.args[command] if command in self.args else {}
        self.args[command] = {**self.args[command], **args}

    def get_name(self):
        """Return the Scenario name"""
        return self.name

    def get_path(self):
        """Return the Scenario path"""
        return self.path

    def is_alive(self):
        """True if Scenario process is running, False otherwise"""
        if hasattr(self, "process") and hasattr(self.process, "process"):
            return self.process.process.is_alive()
        else:
            return False






from .. import defaults

from shapely.wkt import loads as wktToLineString

import geopandas as gpd
import pandas as pd
import numpy as np
import itertools
import logging

class RouteComparison():
    """
    Compare route differences between scenarios

    Parameters
    ----------
    origins: POIDataset
        Points to be used as the origins in route comparison
    dests: POIDataset
        Points to be used as the destinations in route comparison
    origins_id_col: str, optional
        Column in `origins` dataset to be used as ID. Must be unique. If not specified, the index will be used
    dests_id_col: str, optional
        Column in `dests` dataset to be used as ID. Must be unique. If not specified, the index will be used
    cache: bool, optional
        Whether to cache time matrices
    """

    def __init__(self, origins, dests, origins_id_col=None, dests_id_col=None, cache=True):

        ## Set POIDataset indexes to id cols if specified, otherwise use row index
        self.origins = origins.set_index(origins_id_col) if origins_id_col else origins
        self.dests = dests.set_index(dests_id_col) if dests_id_col else dests

        ## Get a DataFrame of all origin_id:dest_id pairs; od=origin:dest
        self.od_pairs = self._get_index_pairs(self.origins, self.dests) \
                            .rename(columns={0: "origin_id", 1: "dest_id"})

        ## Set up cache
        self.cache = cache
        self.duration_matrix = {}

    ##
    ## Base Functions

    def get_duration_matrix(self, scenario, melted=False):
        """
        Calculate a duration matrix between `origins` and `dests`

        Parameters
        ----------
        scenario: Scenario
            Scenario to run calculate the duration matrix on
        melted: bool, optional
            Whether the matrix should be returned as a matrix DF, or melted into `origin_id` `dest_id` and `duration` columns

        Returns
        -------
        pd.DataFrame
            Duration Matrix DataFrame
        """

        s = scenario.get_name()

        ## If hasn't been computed or cache is off
        if s not in self.duration_matrix:

            origin_ids = self.origins.index.tolist()
            origin_coords = self.origins.geometry.apply(lambda o: (o.x, o.y)).tolist()

            dest_ids = self.dests.index.tolist()
            dest_coords = self.dests.geometry.apply(lambda d: (d.x, d.y)).tolist()

            (matrix, origin_snap, dest_snap) \
                = scenario.table(origin_coords,
                                 coords_dest = dest_coords,
                                 output = "np")

            ## Transform matrix to proper dataframe with index&columns
            matrix = pd.DataFrame(matrix, columns = dest_ids, index = origin_ids)

            if self.cache:
                self.duration_matrix[s] = matrix
            else:
                return matrix

        return self.duration_matrix[s] if not melted else self._melt_duration_matrix(self.duration_matrix[s], scenario)

    def get_routes(self, scenario, od_pairs=None):
        """
        Get routes between origins and dests

        Parameters
        ----------
        scenario: Scenario
            Scenario on which the origin:dest routes will be calculated
        od_pairs: pd.DataFrame, optional
            DataFrame of origin:dest pairs in two columns `origin_id` and `dest_id` for calculation. If not specified, all pairwise origin:dest routes will be calculated.

        Returns
        -------
        GeoDataFrame
           GDF with duration, distance, and route geometry for each origin:dest pair
        """

        s = scenario.get_name()

        origins_geom = self.origins.geometry.apply(lambda o: (o.x, o.y)).rename("origin_geom")
        dests_geom = self.dests.geometry.apply(lambda d: (d.x, d.y)).rename("dest_geom")

        ## If origins_dests pairs are not given, use all 
        if od_pairs is None:
            od_pairs = self.od_pairs

        ## Assemble dataframe with `origin_id`, `dest_id`, `origin_geom`, `dest_geom`
        routes = od_pairs.merge(origins_geom.to_frame(), left_on="origin_id",
                                right_index=True) \
                         .merge(dests_geom.to_frame(), left_on="dest_id",
                                right_index=True)

        ## Calculate an OSRM simple_route for a given diff_routes row
        def route_from_row(api, row):
            route = api.simple_route(row["origin_geom"], row["dest_geom"],
                                     geometry = 'wkt', overview="full")

            r_index = ["{}_duration".format(s), "{}_distance".format(s), "geometry"]

            return pd.Series([route['routes'][0]['duration'],
                              route['routes'][0]['distance'],
                              wktToLineString(route["routes"][0]["geometry"])],
                             index=r_index)

        route_info = routes.apply(lambda r: route_from_row(scenario, r),
                                  result_type="expand", axis=1)

        ## Drop point geoms and scenario duration from routes DF (if exists. it will be added
        ## by route_info anyway). Then, join to route info returned, then return GDF
        routes.drop(columns=['origin_geom', 'dest_geom', '{}_duration'.format(scenario.get_name())],
                    inplace=True, errors="ignore")

        return gpd.GeoDataFrame(routes.join(route_info))

    def get_duration_table(self, *args):
        """
        Get route duration table for multiple scenarios

        Parameters
        ----------
        *args
            Arbitrary number of scenarios to run duration table

        Returns
        -------
        DataFrame
            DF of origin:dest pairs and durations for each scenario
        """

        table = self.get_duration_matrix(args[0], melted=True)

        for s_idx in range(1, len(args)):
            table = table.merge(self.get_duration_matrix(args[s_idx], melted=True))

        return table

    ##
    ## Higher level

    def get_difference(self, scenario0, scenario1):
        """Get DF of origin:dest pairs whos routes differ between scenarios"""
        duration_table = self.get_duration_table(scenario0, scenario1)

        s0_duration = duration_table['{}_duration'.format(scenario0.get_name())]
        s1_duration = duration_table['{}_duration'.format(scenario1.get_name())]

        return duration_table[s0_duration != s1_duration]

    def get_same(self, scenario0, scenario1):
        """Get DF of origin:dest pairs whos routes are the same between scenarios"""
        duration_table = self.get_duration_table(scenario0, scenario1)

        s0_duration = duration_table['{}_duration'.format(scenario0.get_name())]
        s1_duration = duration_table['{}_duration'.format(scenario1.get_name())]

        return duration_table[s0_duration == s1_duration]

    ##
    ## Utils

    def _get_index_pairs(self, df0, df1):
        df0_ids = df0.index.tolist()
        df1_ids = df1.index.tolist()

        return pd.DataFrame(list(itertools.product(df0_ids, df1_ids)))

    def _melt_duration_matrix(self, matrix, scenario):
        """Melt a duration matrix into an origin_id:dest_id DF with duration"""

        index_rename = {"level_0": "origin_id",
                        "level_1": "dest_id"}

        return matrix.stack().reset_index() \
                             .rename(columns={
                                 **index_rename,
                                 **{0: "{}_duration".format(scenario.get_name())}
                             })



#!/usr/bin/env python3

import overpass
import hashlib
import logging

from pathlib import Path

from .utils import hash_
from . import defaults

class OSMDataset():
    '''
    OSM data file from which a route network will be extracted

    Parameters
    ----------
    osm_path:
        Path to *.osm{.pbf} dataset
    name: str, optional
        Name of OSMDataset. If not provided, the .osm filename is used.
    '''

    def __init__(self, osm_path, name=None, **kwargs):

        self.log = logging.getLogger(defaults.LOGGER)
        self.path = Path(osm_path)

        ## Check path exists
        if not self.path.is_file():
            raise FileNotFoundError("OSMDataset Not Found ({})".format(osm_path))

        ## Set profile name to filename if not specified
        self.name = name if name else self.path.stem.split(".")[0]

    def get_name(self):
        """Return route network name"""
        return self.name

    def get_path(self):
        """Return route network path"""
        return self.path

    @classmethod
    def from_overpass(cls, query, name=None, overwrite=False, tmp_dir=defaults.TMP_DIR, **kwargs):
        '''
        Initialize an OSMDataset by downloading result of an overpass query and saving as .osm

        Parameters
        ----------
        query : str
            Query to be sent to overpass API. This query should *not* include an `out` directive (eg. [out:xml];)
        name : str
            Name of the route network
        overwrite : bool
            Overwrite route network if it already exists on disk
        tmp_dir : str
            Temporary directory to save route network

        Returns
        -------
        OSMDataset
        '''

        logger = logging.getLogger(defaults.LOGGER)

        ## Use md5 hash of query as filename if name not specified
        out_folder = Path(tmp_dir)
        out_name = name if name else hash_(query)
        out_file = out_folder / "{}.osm".format(out_name)

        ## Honor overwrite settings
        if out_file.is_file():
            if overwrite:
                logger.info("Overwriting {}".format(out_file))
                out_file.unlink()
            else:
                logger.info("Using existing OSMDataset {}".format(out_file))
                return cls(out_file, name=name, overwrite=overwrite, tmp_dir=tmp_dir)

        logger.info("Downloading OSMDataset {}".format(name))

        ## Query API
        oapi = overpass.API()
        xml = oapi.get("[out:xml];{}".format(query),
                       verbosity="geom",
                       responseformat="xml",
                       build=False)

        ## Save file
        out_folder.mkdir(parents=True, exist_ok=True)
        out_file.write_text(xml)

        return cls(out_file, name=name, overwrite=overwrite, tmp_dir=tmp_dir)



from .. import defaults
import logging

class ParallelScenarios():
    """
    Context manager for serving multiple scenarios in parallel. This can be useful for
    comparing routing analysis from multiple scenarios side by side.

    Parameters
    ----------
    *args
        Arbitrary number of scenarios to be compiled, then executed in parallel
    """
    def __init__(self, *args):

        self.log = logging.getLogger(defaults.LOGGER)
        self.scenarios = {scenario.get_name(): scenario for scenario in args}

    def __enter__(self):
        for scenario in self.scenarios.values():
            scenario().__enter__()

        return self.scenarios

    def __exit__(self, exc_type, exc_msg, traceback):
        for scenario in self.scenarios.values():
            scenario.__exit__(exc_type, exc_msg, traceback)



from .. import RoutingProfile
from pathlib import Path

profiles_dir = Path(__file__).parent

bicycle = RoutingProfile(profiles_dir / "bicycle.lua", default=True)
foot = RoutingProfile(profiles_dir / "foot.lua", default=True)
car = RoutingProfile(profiles_dir / "car.lua", default=True)




#!/usr/bin/env python3

from .RoutingProfile import RoutingProfile
from .Environment import Environment
from .OSMDataset import OSMDataset
from .POIDataset import POIDataset
from .Scenario import Scenario

from . import defaults

##
## Logger configuration
##

from logging.config import dictConfig

dictConfig({
    "version": 1,
    'formatters': {
        'f': {
            'format': '[%(levelname)7s] %(message)s',
            'datefmt': '%H:%M:%S'
        }
    },
    'handlers': {
        'h': {
            'class': 'logging.StreamHandler',
            'formatter': 'f',
            'level': 0
        }
    },
    'loggers': {
        defaults.LOGGER: {
        'handlers': ['h'],
        'level': defaults.LOGGER_LEVEL
        }
    },
})



#!/usr/bin/env python3

import shutil
import socket
import hashlib

def hash_(_str):
    return hashlib.md5(_str.encode()).hexdigest()

def find_open_port():
    # Thanks to this gist! https://gist.github.com/jdavis/4040223

    sock = socket.socket()
    sock.bind(('', 0))

    _, port = sock.getsockname()

    return port



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tebetebe",
    version="0.0.1",
    author="1papaya",
    author_email="1love1papaya@gmail.com",
    description="routing analysis with OSRM",
    long_description_content_type="text/markdown",
    long_description="routing analysis with OSRM",
    url="https://github.com/1papaya/tebetebe",
    packages=setuptools.find_packages(),
    include_package_data=True,
    license="MIT",
    install_requires=[
        "osrm",
        "geojson",
        "geopandas",
        "overpass",
        "sh"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.6',
)



#!/usr/bin/env python3

import logging

from pathlib import Path
from . import defaults

class RoutingProfile():
    '''
    A RoutingProfile is a configuration script which represents a routing behaviour, such
    as for bike or car routing. It describes whether or not to traverse a particular type
    of way or node in OSM data, and the speed at which those elements are traversed.

    Check out the osrm-backend wiki for more information!
    https://github.com/Project-OSRM/osrm-backend/wiki/Profiles

    Parameters
    ----------
    lua_path: str
        Path to .lua configuration script
    name: str, optional
        Name of routing profile. If not provided, the .lua filename is used.
    default: bool, optional
        If RoutingProfile is a default profile
    '''

    def __init__(self, lua_path, name=None, default=False, **kwargs):
        self.log = logging.getLogger(defaults.LOGGER)
        self.path = Path(lua_path)
        self.default = default

        ## Check path exists
        if not self.path.is_file():
            raise FileNotFoundError("Routing Profile Not Found ({})".format(lua_path))

        ## Set profile name to lua filename if not specified
        self.name = name if name else self.path.stem

    def is_default(self):
        """Return if routing profile is a default or not"""
        return self.default

    def get_name(self):
        """Return name of routing profile"""
        return self.name

    def get_path(self):
        """Return path of routing profile"""
        return self.path



#!/usr/bin/env python3

from .RoutingProfile import RoutingProfile
from .OSMDataset import OSMDataset
from .POIDataset import POIDataset
from .Scenario import Scenario

from functools import partial
from pathlib import Path

class Environment():
    """
    Environment is a convenience class to set default options for the `tebetebe` base classes.

    Many of the `tebetebe` operations require writing a file to disk in order to be passed
    to the OSRM executables, and it's useful to set a uniform `tmp_dir` for a particular set of
    commands so that all of these files generated by those commands are contained in the same place.

    In addition to `tmp_dir`, any **kwarg passed to Environment will be passed to all of the
    classes available under the Environment. Useful examples of these kwargs include `verbose`
    and `overwrite`

    Any **kwargs set in Environment classes will be overridden if the argument is specified during
    subclass initialization.

    Parameters
    ----------
    **kwargs
        Arbitrary keyword arguments to be passed to each class available under the Environment
    """
    def __init__(self, **kwargs):

        ## Set all env options to this class (set tmp_dir as a Path)
        for k_name, kwarg in kwargs.items():
            setattr(self, k_name, Path(kwarg) if k_name == "tmp_dir" else kwarg)

        self.RoutingProfile = partial(RoutingProfile, **kwargs)
        self.OSMDataset = partial(OSMDataset, **kwargs)
        self.POIDataset = partial(POIDataset, **kwargs)
        self.Scenario = partial(Scenario, **kwargs)

        self.OSMDataset.from_overpass = partial(OSMDataset.from_overpass, **kwargs)
        self.POIDataset.from_overpass = partial(POIDataset.from_overpass, **kwargs)
